from __future__ import annotations

import abc
import dataclasses
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from .collector import Collector
from .types import ParamType

T = TypeVar("T", covariant=True)  #: Item type

W = TypeVar("W", covariant=True)  #: Wrapped item type

if TYPE_CHECKING:
    from .config import Config


class AutoName(Enum):
    """
    Describes automatic handling of an argument name
    """

    #: The argument should not be present in the corresponding source
    FORBIDDEN = 0

    #: Derives the argument name from the original Python identifier (default)
    DERIVED = 1


ArgName = Union[str, AutoName]


@dataclass(frozen=True)
class Arg(abc.ABC):
    """
    Base class for all kinds of arguments
    """

    help: Optional[str] = None  #: Help for the argument

    #: Short option name, used in command line parsing, prefixed by a single hyphen
    short_flag_name: Optional[str] = None

    #: Long option name used in command line argument parsing
    #:
    #: It is lowercase, prefixed with ``--`` and words are separated by hyphens
    long_flag_name: ArgName = AutoName.DERIVED

    def all_flags(self) -> Sequence[str]:
        """
        Returns a sequence of all forms of command line flags
        """
        res: List[str] = []
        if self.short_flag_name is not None:
            res.append(self.short_flag_name)
        assert self.long_flag_name != AutoName.DERIVED
        if isinstance(self.long_flag_name, str):
            res.append(self.long_flag_name)
        return res

    def updated_dict_(self, name: str, help: str, env_prefix: Optional[str]) -> Mapping[str, Any]:
        """
        Returns updated values for this argument, used during :class:`.App` construction

        Args:
            name: Argument field name
            help: Argument docstring which describes the argument role
            env_prefix: Uppercase prefix for all environment variables

        Returns:

        """
        res = {"name": name, "help": help}
        if self.long_flag_name == AutoName.DERIVED:
            res["long_flag_name"] = "--" + name.replace("_", "-")
        return res

    def updated(self, name: str, help: str, env_prefix: Optional[str]) -> Arg:
        return dataclasses.replace(self, **self.updated_dict_(name, help, env_prefix))

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        """
        Returns the keyword arguments for use with argparse.ArgumentParser.add_argument

        Returns:
            Keyword arguments mapping
        """
        return {"help": self.help}


@dataclass(frozen=True)
class Expander(Arg):
    """
    Command-line argument that expands into a flag/value pair
    """

    new_flag: str = ""  #: Inserted flag in the command line
    new_value: str = ""  #: Inserted value in the command line

    def inserts(self) -> Tuple[str, str]:
        """
        Returns the flag/value pair that is inserted when this command flag is present
        """
        return (self.new_flag, self.new_value)

    def updated(self, name: str, help: str, env_prefix: Optional[str]) -> Expander:
        return dataclasses.replace(self, **self.updated_dict_(name, help, env_prefix))

    @staticmethod
    def cmd(
        new_flag: str,
        new_value: str,
        *,
        short_flag_name: Optional[str],
        long_flag_name: ArgName = AutoName.DERIVED,
    ) -> Expander:
        """
        Constructs an expander that inserts a flag/value pair in the command line

        At least one of ``short_flag_name`` or ``long_flag_name`` must be defined.

        Args:
            new_flag: Inserted flag, including the hyphen prefix
            new_value: String value to insert following the flag
            short_flag_name: Short flag name of this command flag
            long_flag_name: Long flag name of this command flag
        """
        res = Expander(
            new_flag=new_flag,
            new_value=new_value,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
        )
        assert res.all_flags(), "Provide at least one of short_flag_name or long_flag_name"
        return res


class Positional(Enum):
    """
    Describes the positional behavior of a parameter
    """

    FORBIDDEN = 0  #: The argument is not positional
    ONCE = 1  #: The argument parses a single positional value
    ZERO_OR_MORE = 2  #: The argument parses the remaining positional value
    ONE_OR_MORE = 3  #: The argument parses at least one remaining positional value

    def should_be_last(self) -> bool:
        """
        Returns whether a positional parameter should be the last one
        """
        return self in {Positional.ZERO_OR_MORE, Positional.ONE_OR_MORE}

    def is_positional(self) -> bool:
        """
        Returns whether a parameter is positional
        """
        return self != Positional.FORBIDDEN


@dataclass(frozen=True)
class Param(Arg, Generic[T]):
    """
    Describes an argument holding a value of a given type

    .. note::
        Instances of :class:`.Param` have two "states":

        * Initially, instances of :class:`.Param` are assigned to class attributes of
        subclasses of :class:`.app.App`. In that state, :attr:`.Param.name` is not set,
        and the other ``XXX_name`` attributes contain either a custom name provided by the user, or
        instructions about the derivation of the corresponding name.

        * When an instance of :class:`.App` is constructed, the :attr:`.name` attribute and the
          ``XXX_name`` attributes of the instance are updated.
    """

    #: Argument type, parser from string to value
    param_type: ParamType[T] = ParamType.invalid()

    #: Argument collector
    collector: Collector[T] = Collector.invalid()  # type: ignore

    default_value: Optional[str] = None  #: Default value inserted as instance

    name: Optional[str] = None  #: Python identifier representing the argument

    positional: Positional = Positional.FORBIDDEN

    #: Configuration key name used in INI files
    #:
    #: It is lowercase, and words are separated by hyphens.
    config_key_name: ArgName = AutoName.DERIVED

    #: Environment variable name
    #:
    #: The environment variable name has an optional prefix, followed by the
    #: Python identifier in uppercase, with underscore as separator.
    #:
    #: This prefix is provided by :attr:`.App.env_prefix_`
    #:
    #: If a non-empty prefix is given, the name is prefixed with it
    #: (and an underscore).
    env_var_name: ArgName = AutoName.FORBIDDEN

    def __call__(self, config: Config) -> T:
        assert self.name is not None, "Needs a constructed App instance to read values from"
        return cast(T, config.values[self.name])

    def updated(self, name: str, help: str, env_prefix: Optional[str]) -> Param[T]:
        return dataclasses.replace(self, **self.replacements(name, help, env_prefix))

    def replacements(self, name: str, help: str, env_prefix: Optional[str]) -> Mapping[str, Any]:
        r = dict(super().updated_dict_(name, help, env_prefix))
        if self.config_key_name == AutoName.DERIVED:
            r["config_key_name"] = name.replace("_", "-")
        if self.env_var_name == AutoName.DERIVED and env_prefix is not None:
            if env_prefix:
                r["env_var_name"] = env_prefix + "_" + name.upper()
            else:
                r["env_var_name"] = name.upper()
        return r

    def all_config_key_names(self) -> Sequence[str]:
        """
        Returns a sequence of all forms of command line options

        Returns:
            Command line options
        """
        if isinstance(self.config_key_name, str):
            return [self.config_key_name]
        else:
            return []

    def all_env_var_names(self) -> Sequence[str]:
        """
        Returns a sequence of all forms of command line options

        Returns:
            Command line options
        """
        if isinstance(self.env_var_name, str):
            return [self.env_var_name]
        else:
            return []

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        res = super().argparse_argument_kwargs()
        if self.default_value is None and self.collector.arg_required():
            res = {**res, "required": True}
        return {
            **res,
            **self.collector.argparse_argument_kwargs(),
            **self.param_type.argparse_argument_kwargs(),
        }

    @staticmethod
    def store(
        param_type: ParamType[T],
        *,
        default_value: Optional[str] = None,
        positional: Positional = Positional.FORBIDDEN,
        short_flag_name: Optional[str] = None,
        long_flag_name: ArgName = AutoName.DERIVED,
        config_key_name: ArgName = AutoName.DERIVED,
        env_var_name: ArgName = AutoName.FORBIDDEN,
    ) -> Param[T]:
        """
        Creates a parameter that stores the last provided value

        If a default value is provided, the argument can be omitted. However,
        if the default_value ``None`` is given (default), then
        the parameter cannot be omitted.

        Args:
            param_type: Parser that transforms a string into a value
            default_value: Default value
            positional: Whether this parameter is present in positional arguments
            short_flag_name: Short option name (optional)
            long_flag_name: Long option name (auto. derived from fieldname by default)
            config_key_name: Config key name (auto. derived from fieldname by default)
            env_var_name: Environment variable name (forbidden by default)

        Returns:
            The constructed Param instance
        """

        return Param(
            param_type=param_type,
            collector=Collector.keep_last(),
            default_value=default_value,
            positional=positional,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
            config_key_name=config_key_name,
            env_var_name=env_var_name,
        )

    @staticmethod
    def append(
        param_type: ParamType[Sequence[W]],
        *,
        positional: Positional = Positional.FORBIDDEN,
        short_flag_name: Optional[str] = None,
        long_flag_name: ArgName = AutoName.DERIVED,
        config_key_name: ArgName = AutoName.DERIVED,
        env_var_name: ArgName = AutoName.FORBIDDEN,
    ) -> Param[Sequence[W]]:
        """
        Creates an argument that stores the last provided value

        Args:
            param_type: Parser that transforms a string into a value
            positional: Whether this argument is present in positional arguments
            short_flag_name: Short option name (optional)
            long_flag_name: Long option name (auto. derived from fieldname by default)
            config_key_name: Config key name (auto. derived from fieldname by default)
            env_var_name: Environment variable name (forbidden by default)

        Returns:
            The constructed Arg instance
        """
        return Param(
            param_type=param_type,
            collector=Collector.append(),
            default_value=None,
            positional=positional,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
            config_key_name=config_key_name,
            env_var_name=env_var_name,
        )



from __future__ import annotations

import abc
import argparse
import os
import sys
import textwrap
import typing
from cmath import exp
from collections import OrderedDict, _OrderedDictItemsView
from configparser import ConfigParser
from dataclasses import dataclass
from distutils.cmd import Command
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
)
from typing import OrderedDict as OrderedDictT
from typing import Sequence, Tuple, Type, TypeVar, Union, cast

import class_doc

from . import sources, types
from .arg import Arg, Expander, Param, Positional
from .collector import Instance
from .errors import ArgErr, Err, ManyErr, Result, collect, map_result
from .sources import CommandLine, EnvironmentVariables, IniSectionSource, Source
from .util import dict_from_multiple_keys, filter_ordered_dict_by_value_type

if TYPE_CHECKING:
    from .app import App

A = TypeVar("A", bound="App")
T = TypeVar("T", covariant=True)


@dataclass(frozen=True)
class Config:
    """
    Holds parsed parameter values
    """

    values: Mapping[str, Any]

    def __getitem__(self, param: Param[T]) -> T:
        assert param.name is not None
        return cast(T, self.values[param.name])

    @staticmethod
    def make(
        app: App,
        cwd: Path,
        args: Sequence[str],
        env: Mapping[str, str],
    ) -> Result[Config]:
        return Processor(app, cwd, args, env).process()


@dataclass(frozen=True)
class Processor:
    """
    Processes environment variables, configuration files and command line arguments
    """

    app: App  #: App describing the configuration arguments
    cwd: Path  #: Path from which relative paths for conf. files are resolved
    args: Sequence[str]  #: Command line arguments as a sequence of strings
    env: Mapping[str, str]  #: Environment variables

    def command_line(self) -> Result[CommandLine]:
        """
        Constructs a command line source

        Returns:
            A source or an error
        """
        exp_flags = {flag: cmd.inserts() for (flag, cmd) in self.app.args_.cl_expanders.items()}
        val_flags = frozenset(self.app.args_.cl_params.keys())
        return CommandLine.make(self.args, exp_flags, val_flags)

    def environment_variables(self) -> Result[EnvironmentVariables]:
        """
        Constructs a source from environment variables

        Returns:
            A source or an error
        """
        return EnvironmentVariables(self.env)

    def process(self) -> Result[Config]:
        """
        Runs the processing

        Returns:
            A parsed configuration or an error
        """
        app = self.app
        config_sources: Result[Tuple[CommandLine, EnvironmentVariables]] = collect(
            self.command_line(), self.environment_variables()
        )
        if isinstance(config_sources, Err):
            return config_sources
        cl, env = config_sources
        ini_from_env: Result[Sequence[Instance[Sequence[Path]]]] = env[app.ini_files]
        if isinstance(ini_from_env, Err):
            return ini_from_env
        ini_from_cl: Result[Sequence[Instance[Sequence[Path]]]] = cl[app.ini_files]
        if isinstance(ini_from_cl, Err):
            return ini_from_cl
        sources: List[Source] = [env]
        ini_instances: Sequence[Path] = [
            ins for instances in [*ini_from_env, *ini_from_cl] for ins in instances.value
        ]
        for fn in ini_instances:
            sections_res = IniSectionSource.from_file(
                self.cwd / fn, app.ini_sections_(), app.args_.cf_params
            )
            if isinstance(sections_res, Err):
                return sections_res
            sources.extend(sections_res)
        sources.append(cl)
        values: List[Tuple[str, Any]] = []
        errs: List[Err] = []
        for name, arg in app.args_.params.items():
            res = Source.collect(sources, arg)
            if isinstance(res, Err):
                errs.append(ArgErr(arg, res))
            else:
                values.append((name, res))
        if errs:
            return ManyErr(errs)
        else:
            return Config(dict(values))



from __future__ import annotations

import pathlib
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Literal,
    Mapping,
    NoReturn,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import parsy

from configpile.util import assert_never

from .errors import ParseErr, Result, collect_seq

T = TypeVar("T", covariant=True)  #: Item type

W = TypeVar("W", covariant=True)  #: Wrapped item type


class ForceCase(Enum):
    """
    Describes whether a string is normalized to lower or upper case before processing
    """

    NO_CHANGE = 0  #: Keep case
    UPPER = 1  #: Change to uppercase
    LOWER = 2  #: Change to lowercase


class ParamType(ABC, Generic[T]):
    """Describes a parameter type"""

    @abstractmethod
    def parse(self, arg: str) -> Result[T]:
        """
        Parses a string into a result

        This method reports parsing errors using a result type instead of raising
        exceptions.

        Args:
            arg: String value to parse

        Returns:
            A result containing either the parsed value or a description of an error
        """

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        return {}

    def empty_means_none(self, strip: bool = True) -> ParamType[Optional[T]]:
        """
        Returns a new parameter type where the empty string means None

        Args:
            strip: Whether to strip whitespace

        Returns:
            A new parameter type
        """
        return _EmptyMeansNone(self, strip)

    def separated_by(
        self, sep: str, strip: bool = True, remove_empty: bool = True
    ) -> ParamType[Sequence[T]]:
        """
        Returns a new parameter type that parses values separated by a string

        Args:
            sep: Separator
            strip: Whether to strip whitespace from separated values
            remove_empty: Whether to remove empty strings before parsing them

        Returns:
            A new parameter type
        """
        return _SeparatedBy(self, sep, strip, remove_empty)

    @staticmethod
    def from_parser(type_: Type[T], parser: parsy.Parser) -> ParamType[T]:
        """
        Creates a parameter type from a parsy parser

        Args:
            type_: PEP 484 type, used to type the return argument
            parser: Parser returning a value of type ``t``

        Returns:
            Parameter type
        """
        return _Parsy(parser)

    @staticmethod
    def from_function_that_raises(f: Callable[[str], T]) -> ParamType[T]:
        """
        Creates a parameter type from a function that raises exceptions on parse errors

        Args:
            f: Function that parses the string

        Returns:
            Parameter type
        """
        return _FunctionThatRaises(f)

    @staticmethod
    def from_result_function(f: Callable[[str], Result[T]]) -> ParamType[T]:
        """
        Creates a parameter type from a function that returns a value or an error

        Args:
            f: Function that parses the string

        Returns:
            Parameter type
        """

        return _ResultFunction(f)

    @staticmethod
    def invalid() -> ParamType[NoReturn]:
        """
        Creates a parameter type that always return errors
        """

        def invalid_fun(s: str) -> NoReturn:
            raise RuntimeError("Invalid parameter type")

        return ParamType.from_function_that_raises(invalid_fun)

    @staticmethod
    def choices_str(
        values: Iterable[str],
        strip: bool = True,
        force_case: ForceCase = ForceCase.NO_CHANGE,
    ) -> ParamType[str]:
        """
        Creates a parameter type whose values are chosen from a set of strings

        Args:
            values: Set of values
            strip: Whether to strip whitespace before looking for choices

        Returns:
            Parameter type
        """

        return ParamType.choices({v: v for v in values}, strip, force_case)

    @staticmethod
    def choices(
        mapping: Mapping[str, T],
        strip: bool = True,
        force_case: ForceCase = ForceCase.NO_CHANGE,
    ) -> ParamType[T]:
        """
        Creates a parameter type whose strings correspond to keys in a dictionary

        Args:
            mapping: Dictionary mapping strings to values
            strip: Whether to strip whitespace before looking for keys

        Returns:
            Parameter type
        """
        return _Choices(mapping, strip, force_case)


@dataclass(frozen=True)
class _Choices(ParamType[T]):
    """
    Describes a multiple choice parameter type
    """

    mapping: Mapping[str, T]
    strip: bool
    force_case: ForceCase = ForceCase.NO_CHANGE

    def parse(self, arg: str) -> Result[T]:
        if self.strip:
            arg = arg.strip()
        if self.force_case is ForceCase.LOWER:
            arg = arg.lower()
        elif self.force_case is ForceCase.UPPER:
            arg = arg.upper()
        elif self.force_case is ForceCase.NO_CHANGE:
            pass
        else:
            assert_never(self.force_case)
        if arg in self.mapping:
            return self.mapping[arg]
        else:
            msg = f"Value {arg} not in choices {','.join(self.mapping.keys())}"
            return ParseErr(msg, arg, None)

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        return {"choices": self.mapping.keys(), "type": str}


@dataclass  # not frozen because mypy bug, please be responsible
class _FunctionThatRaises(ParamType[T]):
    """
    Wraps a function that may raise exceptions
    """

    # the optional is to make mypy happy
    fun: Callable[[str], T]  #: Callable function that may raise

    def parse(self, arg: str) -> Result[T]:
        try:
            f = self.fun
            assert f is not None
            return f(arg)
        except Exception as err:
            return ParseErr(str(err), arg, None)


@dataclass  # not frozen because mypy bug, please be responsible
class _ResultFunction(ParamType[T]):
    """
    Wraps a function that returns a result
    """

    fun: Callable[[str], Result[T]]

    def parse(self, arg: str) -> Result[T]:
        return self.fun(arg)


@dataclass(frozen=True)
class _Parsy(ParamType[T]):
    """
    Wraps a parser from the parsy library
    """

    parser: parsy.Parser

    def parse(self, arg: str) -> Result[T]:
        res = (self.parser << parsy.eof)(arg, 0)  # Inspired by Parser.parse
        if res.status:
            return cast(T, res.value)
        else:
            return ParseErr(res.expected, arg, res.furthest)


@dataclass(frozen=True)
class _EmptyMeansNone(ParamType[Optional[W]]):
    """
    Wraps an existing parameter type so that "empty means none"
    """

    wrapped: ParamType[W]  #: Wrapped ParamType called if value is not empty
    strip: bool  #:  Whether to strip whitespace before testing for empty

    def parse(self, value: str) -> Result[Optional[W]]:
        if self.strip:
            value = value.strip()
        if not value:
            return None
        else:
            return self.wrapped.parse(value)


@dataclass(frozen=True)
class _SeparatedBy(ParamType[Sequence[W]]):
    """
    Parses values separated by a given separator
    """

    item: ParamType[W]  #: ParamType of individual items
    sep: str  #: Item separator
    strip: bool  #: Whether to strip whitespace around separated strings
    remove_empty: bool  #: Whether to prune empty strings

    def parse(self, arg: str) -> Result[Sequence[W]]:
        items: Iterable[str] = arg.split(self.sep)
        if self.strip:
            items = map(lambda s: s.strip(), items)
        if self.remove_empty:
            items = filter(None, items)
        res: Sequence[Result[W]] = [self.item.parse(s) for s in items]
        return collect_seq(res)


path = ParamType.from_function_that_raises(lambda s: pathlib.Path(s))
integer = ParamType.from_function_that_raises(lambda s: int(s))
word = ParamType.from_function_that_raises(lambda s: s.strip())



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

if TYPE_CHECKING:
    from .arg import Param


class Err(ABC):
    """
    Describes an error that occurred during argument parsing
    """

    @abstractmethod
    def markdown(self) -> Sequence[str]:
        pass

    @abstractmethod
    def errors(self) -> Sequence[Err]:
        pass


class SingleErr(Err):
    def errors(self) -> Sequence[Err]:
        return [self]


@dataclass
class GenErr(SingleErr):
    msg: str

    def markdown(self) -> Sequence[str]:
        return [self.msg]


@dataclass
class ArgErr(SingleErr):
    arg: Param[Any]
    err: Err

    def markdown(self) -> Sequence[str]:
        return [f"Error while processing argument {self.arg.name}:", "", *self.err.markdown()]


@dataclass
class ManyErr(Err):

    errs: Sequence[Err]

    def __post_init__(self) -> None:
        assert len(self.errs) > 0, "A ManyErr should contain at least one error"

    def markdown(self) -> Sequence[str]:
        lines: List[str] = []
        for i, e in enumerate(self.errs):
            start = f"{i+1}. "
            res: Sequence[str] = e.markdown()
            if res:
                line1 = start + res[0]

                def space_prefix(s: str) -> str:
                    return (" " * len(start)) + s

                rest: List[str] = [space_prefix(l) for l in res[1:]]
                lines.append(line1)
                lines.extend(rest)
        return lines

    def errors(self) -> Sequence[Err]:
        return self.errs


@dataclass(frozen=True)
class ParseErr(SingleErr):
    msg: str
    string: str  #: String value that could not be parsed
    index: Optional[int]  #: Position of the parse error, if known

    def markdown(self) -> Sequence[str]:
        parse_msg = f"Parse error in `{self.string}`"
        if self.index is not None:
            parse_msg += f" at index {self.index}"
        return [self.msg, "", parse_msg]


#: Ok value in our custom result type
T = TypeVar("T", covariant=True)

U = TypeVar("U")
V = TypeVar("V")
W = TypeVar("W")


Result = Union[T, Err]


def collect_seq(seq: Sequence[Result[T]]) -> Result[Sequence[T]]:
    ok: List[T] = []
    errs: List[Err] = []
    for res in seq:
        if isinstance(res, Err):
            errs.extend(res.errors())
        else:
            ok.append(res)
    if errs:
        return ManyErr(errs)
    else:
        return ok


@overload
def collect(t: Result[T], u: Result[U]) -> Result[Tuple[T, U]]:
    pass


@overload
def collect(t: Result[T], u: Result[U], v: Result[V]) -> Result[Tuple[T, U, V]]:
    pass


@overload
def collect(
    t: Result[T],
    u: Result[U],
    v: Result[V],
    w: Result[W],
) -> Result[Tuple[T, U, V, W]]:
    pass


def collect(*args):  # type: ignore[no-untyped-def]
    ok: List[Any] = []
    errs: List[Err] = []
    for arg in args:
        if isinstance(arg, Err):
            errs.extend(arg.errors())
        else:
            ok.append(arg)
    if errs:
        return ManyErr(errs)
    else:
        return tuple(ok)


def map_result(f: Callable[[T], U], r: Result[T]) -> Result[U]:
    if isinstance(r, Err):
        return r
    else:
        return f(r)



from __future__ import annotations

import abc
import configparser
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, Generic, Iterable, List, Literal, Mapping, Sequence, Tuple, TypeVar

from .arg import Param, Positional
from .collector import Instance
from .errors import ArgErr, Err, GenErr, ManyErr, Result, collect_seq

T = TypeVar("T")  #: Item type


class Source(abc.ABC):
    """
    Describes a source of argument name/value pairs
    """

    @abstractmethod
    def get_strings(self, param: Param[T]) -> Result[Sequence[str]]:
        """
        Returns a sequence of string values corresponding to the given parameter

        Args:
            param: Parameter to collect the strings for

        Returns:
            A sequence of strings or an error
        """
        pass

    def __getitem__(self, param: Param[T]) -> Result[Sequence[Instance[T]]]:
        """
        Returns instances of all parsed values corresponding to the given parameter

        Args:
            param: Parameter to collect the instance for

        Returns:
            A sequence of instances or an error
        """
        res = self.get_strings(param)
        if isinstance(res, Err):
            return res
        else:
            return collect_seq([Instance.parse(s, param.param_type, source=self) for s in res])

    @staticmethod
    def collect_instances(
        sources: Sequence[Source], param: Param[T]
    ) -> Result[Sequence[Instance[T]]]:
        """
        Parses all instances of a parameter in a sequence of sources

        Args:
            sources: Sequence of sources to get the strings from
            param: Param to parse

        Returns:
            A sequence of instances or an error
        """
        ok: List[Instance[T]] = []
        errs: List[Err] = []
        for source in sources:
            res = source[param]
            if isinstance(res, ManyErr):
                errs.extend(res.errs)
            elif isinstance(res, Err):
                errs.append(res)
            else:
                ok.extend(res)
        if errs:
            return ManyErr(errs)
        else:
            return ok

    @staticmethod
    def collect(sources: Sequence[Source], param: Param[T]) -> Result[T]:
        """
        Parses and collect the value of a parameter from a sequence of sources

        Args:
            sources: Sequence of sources to get the strings from
            param: Parameter to parse and collect

        Returns:
            The collected value or an error
        """
        instances = Source.collect_instances(sources, param)
        if isinstance(instances, Err):
            return instances
        return param.collector.collect(instances)


@dataclass(frozen=True)
class EnvironmentVariables(Source):
    """
    Source coming from environment variables
    """

    env: Mapping[str, str]

    def get_strings(self, arg: Param[T]) -> Result[Sequence[str]]:
        if isinstance(arg.env_var_name, str) and arg.env_var_name in self.env:
            return [self.env[arg.env_var_name]]
        else:
            return []


@dataclass(frozen=True)
class IniSection:
    """
    Describes a section of an INI file to include in the current configuration
    """

    name: str  #: Section name
    strict: bool  #: Whether all the keys must correspond to parsed arguments


@dataclass(frozen=True)
class IniSectionSource(Source):
    """
    Source coming from the section of an INI file
    """

    filename: Path
    section_name: str
    elements: Mapping[str, str]

    def get_strings(self, arg: Param[T]) -> Result[Sequence[str]]:
        if isinstance(arg.config_key_name, str) and arg.config_key_name in self.elements:
            return [self.elements[arg.config_key_name]]
        else:
            return []

    @staticmethod
    def from_file(
        ini_file: Path, sections: Sequence[IniSection], valid_keys: Iterable[str]
    ) -> Result[Sequence[IniSectionSource]]:
        config = configparser.ConfigParser()
        try:
            with open(ini_file, "r") as f:
                config.read_file(f)
        except FileNotFoundError as e:
            return GenErr(f"File {ini_file} not found")
        res: List[IniSectionSource] = []
        for s in sections:  # loop through sections, later section values override earlier ones
            elements: Dict[str, str] = {}
            if s.name in config.sections():
                data: configparser.SectionProxy = config[s.name]
                if s.strict:
                    # we're strict, so we list all keys present in the section and match them
                    for k in data.keys():
                        if k not in valid_keys:
                            return GenErr(f"Invalid key {k} in section {s.name} of {ini_file}")
                        elements[k] = data[k]  # insert value
                else:
                    # we're not strict, so we only extract the valid keys
                    for k in valid_keys:
                        if k in data:
                            elements[k] = data[k]
            if elements:
                res.append(IniSectionSource(ini_file, s.name, elements))
        return res


@dataclass(frozen=True)
class CommandLine(Source):
    """
    Source describing command line arguments
    """

    pairs: Sequence[Tuple[str, str]]  #: Key/value argument pairs
    positional: Sequence[str]  #: Remaining positional values
    commands: Sequence[str]  #: Remaining commands

    def get_strings(self, arg: Param[T]) -> Result[Sequence[str]]:
        from_pairs: List[str] = [value for (key, value) in self.pairs if key in arg.all_flags()]
        from_pos: List[str] = []
        if arg.positional != Positional.FORBIDDEN:
            if arg.positional == Positional.ONCE and len(self.positional) != 1:
                return GenErr("One positional argument must be provided")
            if arg.positional == Positional.ONE_OR_MORE and len(self.positional) == 0:
                return GenErr("At least one positional argument must be provided")
            from_pos = list(self.positional)
        return [*from_pos, *from_pairs]

    @staticmethod
    def make(
        args: Sequence[str],
        expanding_flags: Mapping[str, Tuple[str, str]],
        flags_followed_by_value: Iterable[str],
        command_flags: Iterable[str] = [],
    ) -> Result[CommandLine]:
        """
        Returns processed command line arguments

        Args:
            args: Raw command line arguments
            expanding_flags: Flags that expand to a key/value pair
            flags_followed_by_value: Flags that are followed by a value

        Returns:
            The processed command line
        """
        pairs: List[Tuple[str, str]] = []
        commands: List[str] = []
        rest: List[str] = []
        i = 0
        while i < len(args):
            k = args[i]
            if k in expanding_flags:
                kv = expanding_flags[k]
                pairs.append(kv)
            elif k in flags_followed_by_value:
                i += 1
                if i >= len(args):
                    return GenErr("Unexpected end of command line")
                else:
                    v = args[i]
                    pairs.append((k, v))
            elif k in command_flags:
                commands.append(k)
            else:
                rest.append(k)
            i += 1
        return CommandLine(pairs=pairs, positional=rest, commands=commands)



from __future__ import annotations

import argparse
import os
import shutil
import sys
import textwrap
from cmath import exp
from collections import OrderedDict
from configparser import ConfigParser
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
)
from typing import OrderedDict as OrderedDictT
from typing import Sequence, Tuple, Type, TypeVar

import class_doc

from . import sources, types
from .arg import Arg, Expander, Param
from .config import Config
from .errors import Err
from .util import dict_from_multiple_keys, filter_ordered_dict_by_value_type

A = TypeVar("A", bound="App")
T = TypeVar("T", covariant=True)


@dataclass(frozen=True)
class Args:
    """
    Describes the arguments present in an app
    """

    #: All arguments present in the app
    #:
    #: Those are indexed by the field name
    all: Mapping[str, Arg]

    #: All parameters present in the app, for which a value is recovered
    #:
    #: Those are indexed by the field name
    params: Mapping[str, Param]  # type: ignore[type-arg]

    #: All command line flags that expand into a (key, value) pair of strings
    cl_expanders: Mapping[str, Expander]

    #: All command line arguments that are followed by a value
    cl_params: Mapping[str, Param]  # type: ignore[type-arg]

    #: The positional arguments
    cl_positional: Optional[Param]  # type: ignore[type-arg]

    #: Map of all config file keys
    cf_params: Mapping[str, Param]  # type: ignore[type-arg]

    #: Map of all supported env. variables
    env_params: Mapping[str, Param]  # type: ignore[type-arg]

    @staticmethod
    def populated_base_args(cls: Type[A]) -> OrderedDictT[str, Arg]:
        """
        Returns an ordered dictionary of configuration arguments

        In the returned dictionary, the :class:`.Arg` instances are updated with their
        field name and help string.

        Args:
            cls: App class type

        Returns:
            Argument dictionary
        """
        docs: Mapping[str, Sequence[str]] = class_doc.extract_docs_from_cls_obj(cls)

        def get_help(name: str) -> str:
            seq = docs.get(name, [])
            return textwrap.dedent("\n".join(seq))

        elements: List[Tuple[str, Arg]] = [
            (
                name,
                arg.updated(name=name, help=get_help(name), env_prefix=cls.env_prefix_),
            )
            for parent_including_itself in cls.__mro__
            for (name, arg) in parent_including_itself.__dict__.items()
            if isinstance(arg, Arg) and not name.endswith("_")
        ]
        return OrderedDict(elements)

    @staticmethod
    def from_app_class(cls: Type[A]) -> Args:
        """
        Returns a :class:`.Args` instance populated with updated arguments

        Args:
            cls: The :class:`.App` subclass to investigate

        Returns:
            An object holding the arguments/flags sorted by type
        """
        all: OrderedDictT[str, Arg] = Args.populated_base_args(cls)
        fields: OrderedDictT[str, Param] = filter_ordered_dict_by_value_type(Param, all)  # type: ignore[type-arg]

        cl_all: Mapping[str, Arg] = dict_from_multiple_keys(
            [(arg.all_flags(), arg) for arg in all.values()]
        )
        cl_expanders: Mapping[str, Expander] = {
            k: v for (k, v) in cl_all.items() if isinstance(v, Expander)
        }
        cl_flag_args: Mapping[str, Param] = {k: v for (k, v) in cl_all.items() if isinstance(v, Param)}  # type: ignore[type-arg]
        cl_pos_args: Sequence[Param] = [a for a in fields.values() if a.positional.is_positional()]  # type: ignore[type-arg]
        if len(cl_pos_args) == 0:
            cl_positional = None
        elif len(cl_pos_args) == 1:
            cl_positional = cl_pos_args[0]
        else:
            raise ValueError("At most one positional argument can be given")
        cf_keys: Mapping[str, Param] = dict_from_multiple_keys(  # type: ignore[type-arg]
            [(arg.all_config_key_names(), arg) for arg in fields.values()]
        )
        env_vars: Mapping[str, Param] = dict_from_multiple_keys(  # type: ignore[type-arg]
            [(arg.all_env_var_names(), arg) for arg in fields.values()]
        )
        assert not any(
            map(lambda a: a.positional.should_be_last(), cl_pos_args[:-1])
        ), "Only the last positional argument may take a variable number of values"
        return Args(
            all=all,
            params=fields,
            cl_expanders=cl_expanders,
            cl_params=cl_flag_args,
            cl_positional=cl_positional,
            cf_params=cf_keys,
            env_params=env_vars,
        )


class App:
    """
    A base class for the configuration of Python scripts
    """

    #: Configuration file paths
    #:
    #: The paths are absolute or relative to the current working directory, and
    #: point to existing INI files containing configuration settings
    ini_files: Param[Sequence[Path]] = Param.append(types.path.separated_by(","))

    # TODO help: Optional[HelpCmd] = HelpCmd(short_flag_name="-h")  #: Show a help message and exit

    #: Names of sections to parse in configuration files, with unknown keys ignored
    ini_relaxed_sections_: Sequence[str] = ["Common", "COMMON", "common"]

    #: Names of additional sections to parse in configuration files, unknown keys error
    ini_strict_sections_: Sequence[str] = []

    @classmethod
    def ini_sections_(cls) -> Sequence[sources.IniSection]:
        """
        Returns a sequence of INI file sections to parse

        By default, this parses first the relaxed sections and then the strict ones.

        This method can be overridden.
        """
        relaxed = [sources.IniSection(name, False) for name in cls.ini_relaxed_sections_]
        strict = [sources.IniSection(name, True) for name in cls.ini_strict_sections_]
        return relaxed + strict

    prog_: Optional[str] = None  #: Program name
    description_: Optional[str] = None  #: Text to display before the argument help
    env_prefix_: Optional[str] = None  #: Uppercase prefix of environment variables

    args_: Args  #: Arguments

    @classmethod
    def app_(cls: Type[A]) -> A:
        """
        Creates an instance with updated fields

        This class method should be called on subclasses of :class:`.App`.

        Returns:
            An instance of App
        """
        res = cls()
        assert (
            not cls.ini_files.positional.is_positional()
        ), "Configuration files cannot be given as positional arguments"
        if res.prog_ is None:
            res.prog_ = sys.argv[0]
        if res.description_ is None:
            res.description_ = cls.__doc__
        res.args_ = Args.from_app_class(cls)
        for name, arg in res.args_.all.items():
            res.__setattr__(name, arg)
        return res

    def parse_(
        self,
        cwd: Path = Path.cwd(),
        args: Sequence[str] = sys.argv[1:],
        env: Mapping[str, str] = os.environ,
    ) -> Config:
        """
        Parses multiple information sources into a configuration and display help on error

        Default values are taken from the current working directory, the script command line
        arguments, and the current environment variables.

        Args:
            cwd: Directory used as a base for the configuration file relative paths
            args: Command line arguments
            env: Environment variables

        Returns:
            A parsed configuration
        """
        res = Config.make(self, cwd, args, env)
        if isinstance(res, Err):
            try:
                from rich.console import Console
                from rich.markdown import Markdown

                console = Console()
                md = Markdown("\n".join(res.markdown()))
                console.print(md)
            except:
                sz = shutil.get_terminal_size()
                t = res.markdown()
                print(textwrap.fill("\n".join(t), width=sz.columns))
            self.argument_parser_().print_help()
            sys.exit(1)
        return res

    def argument_parser_(self) -> argparse.ArgumentParser:
        """
        Returns an :class:`argparse.ArgumentParser` for documentation purposes
        """
        p = argparse.ArgumentParser(prog=self.prog_, description=self.description_)
        for arg in self.args_.cl_params.values():
            p.add_argument(
                *arg.all_flags(),
                **arg.argparse_argument_kwargs(),
            )
        return p



from __future__ import annotations

import abc
from dataclasses import dataclass
from enum import Enum
from operator import truediv
from typing import TYPE_CHECKING, Any, Generic, List, Mapping, NoReturn, Sequence, TypeVar

from .errors import GenErr, Result, map_result
from .types import ParamType

T = TypeVar("T")  #: Item type

W = TypeVar("W")  #: Wrapped item type

if TYPE_CHECKING:
    from .sources import Source


@dataclass(frozen=True)
class Instance(Generic[T]):
    """
    Parameter instance
    """

    value: T  #: Value parsed from string
    string: str  #: Parsed string
    source: Source  #: Parameter source

    @staticmethod
    def parse(s: str, param_type: ParamType[T], source: Source) -> Result[Instance[T]]:
        """
        Parses a string using a parameter type and constructs a parameter instance

        Args:
            s: String to parse
            param_type: Parameter type used as a parser
            source: Source of the string

        Returns:
            A parameter instance or an error
        """

        def res_to_instance(v: T) -> Instance[T]:
            return Instance(v, s, source)

        return map_result(res_to_instance, param_type.parse(s))


class Collector(abc.ABC, Generic[T]):
    """
    Collects argument instances and computes the final value
    """

    @abc.abstractmethod
    def arg_required(self) -> bool:
        """
        Returns whether one instance of the argument needs to be present
        """
        pass

    @abc.abstractmethod
    def collect(self, seq: Sequence[Instance[T]]) -> Result[T]:
        pass

    @abc.abstractmethod
    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        pass

    @staticmethod
    def keep_last() -> Collector[T]:
        """
        Returns a collector that keeps the last value
        """
        return _KeepLast()

    @staticmethod
    def append() -> Collector[Sequence[W]]:
        """
        Returns a collector that appends sequences
        """
        return _Append()

    @staticmethod
    def invalid() -> Collector[NoReturn]:
        """
        Returns an invalid collector that always returns an error
        """
        return _Invalid()


class _KeepLast(Collector[T]):
    def arg_required(self) -> bool:
        return True

    def collect(self, seq: Sequence[Instance[T]]) -> Result[T]:
        if not seq:  # no instances provided
            return GenErr("Argument is required")
        else:  # instances are provided
            return seq[-1].value

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        return {"action": "store"}


class _Append(Collector[Sequence[W]]):
    def arg_required(self) -> bool:
        return False

    def collect(self, seq: Sequence[Instance[Sequence[W]]]) -> Result[Sequence[W]]:
        res: List[W] = []
        for i in seq:
            res.extend(i.value)
        return res

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        return {"action": "append"}


class _Invalid(Collector[Any]):
    def arg_required(self) -> bool:
        raise NotImplementedError

    def collect(self, seq: Sequence[Instance[T]]) -> Result[T]:
        return GenErr("Invalid collector")

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        raise NotImplementedError(
            "This should have be replaced by a valid collector during construction"
        )



from collections import OrderedDict
from dataclasses import dataclass
from typing import Callable, Dict, Generic, List, NoReturn, Optional
from typing import OrderedDict as OrderedDictT
from typing import Sequence, Tuple, Type, TypeVar, ValuesView, overload

K = TypeVar("K")
T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
W = TypeVar("W")
E = TypeVar("E")  #: Error type


def dict_from_multiple_keys(pairs: Sequence[Tuple[Sequence[K], V]]) -> Dict[K, V]:
    """
    Constructs a dict from a list of items where a value can have multiple keys

    Args:
        pairs: List of dict elements

    Returns:
        A dictionary
    """
    return {k: v for (kk, v) in pairs for k in kk}


def filter_ordered_dict_by_value_type(w: Type[W], od: OrderedDictT[K, V]) -> OrderedDictT[K, W]:
    """
    Filters values of an ordered dictionary that correspond to a given type

    Returns:
        A new dictionary
    """
    pairs: Sequence[Tuple[K, W]] = [(k, v) for (k, v) in od.items() if isinstance(v, w)]
    return OrderedDict(pairs)


def filter_ordered_dict(
    f: Callable[[K, V], bool],
    od: OrderedDictT[K, V],
) -> OrderedDictT[K, V]:
    """
    Filters items of an ordered dictionary based on a predicate

    Args:
        f: Predicate taking a key/value pair as arguments
        od: Ordered dictionary to filter

    Returns:
        A new dictionary
    """
    return OrderedDict([(k, v) for (k, v) in od.items() if f(k, v)])


def filter_sequence_by_value_type(
    w: Type[W], seq: Sequence[V], predicate: Optional[Callable[[W], bool]]
) -> Sequence[W]:
    """
    Filter values using their type and an optional predicate

    Args:
        w: Type to keep
        seq: Sequence to filter
        predicate: Optional predicate, default: keep all values of given type

    Returns:
        The filtered sequence
    """
    if predicate is None:
        predicate = lambda w: True
    return [v for v in seq if isinstance(v, w) if predicate(v)]


def assert_never(value: NoReturn) -> NoReturn:
    """
    Function used for exhaustiveness checks

    See `<https://github.com/python/cpython/pull/30842>`_
    """
    assert False, f"Unhandled value: {value} ({type(value).__name__})"
